import { Client } from '@elastic/elasticsearch';
import { Config, Init, Provide } from '@midwayjs/core';
import { indices, getType } from './indices';
import { get, toLower } from 'lodash';
import * as dayjs from 'dayjs';

@Provide()
export class ESService {
  @Config('elasticsearch')
  esConfig: object;
  /**es连接实例 */
  esClient: Client;

  @Init()
  async initClient() {
    const esClient = new Client(this.esConfig);
    this.esClient = esClient;
  }
  /**
   * 1.索引(es中的表)建立:根据编写的文件生成
   * 2.初始化数据,通过接口触发
   * 3.对指定表进行的数据修改做同步
   */

  async initIndex() {
    const tables = [];
    for (const tableName in indices) {
      const index = toLower(tableName);
      const hasIndex = await this.esClient.indices.exists({ index });
      const mappings = indices[tableName];
      if (hasIndex) {
        await this.esClient.indices.delete({ index });
      }
      const properties = {};
      for (const key in mappings) {
        let pk = key;
        const val = mappings[key];
        properties[pk] = getType(val);
      }
      await this.esClient.indices.create({ index, mappings: { properties } });
      tables.push(tableName);
    }
    return tables;
  }
  /**
   * 按表,初始化es数据
   * 1.获取数据总量
   * 2.${this.limit}步进循环处理
   * 3.处理特殊字段的数据,然后添加进es中
   * @param tableName 表名,全小写
   * @param data 数据
   * @param mappings 设置的映射
   */
  async initData(tableName, data) {
    const mapping = indices[tableName];
    const datas = await this.dealData(data, mapping);
    if (datas.length <= 0) return;
    await this.addDatas(tableName, datas);
  }

  /**
   * 处理数据
   * @param list model查出来的数据
   * @param mappings indices设置的映射
   */
  async dealData(list, mappings) {
    const datas = [];
    for (const i of list) {
      const obj = {};
      for (const key in mappings) {
        const val = get(i, key);
        if (val) {
          // 检查是否是date类型,如果是date类型需要检查数据是否正确
          const dataType = mappings[key];
          if (dataType === 'date') {
            const isDate = dayjs(val).isValid();
            if (!isDate) continue;
          }
          let pk = key;
          obj[pk] = val;
        }
      }
      const keys = Object.keys(obj);
      if (keys.length > 0) datas.push(obj);
    }
    return datas;
  }

  /**
   * upsert至es
   * @param index 全小写表名
   * @param datas 数据
   */
  async addDatas(index, datas) {
    await this.esClient.helpers.bulk({
      datasource: datas,
      onDocument(doc) {
        return [{ update: { _index: index, _id: get(doc, 'id') } }, { doc_as_upsert: true }];
      },
    });
  }
}
