import { Client } from '@elastic/elasticsearch';
import { Config, Init, Provide } from '@midwayjs/core';
import { floor, get } from 'lodash';

@Provide()
export class SearchService {
  @Config('elasticsearch')
  esConfig: object;
  /**es连接实例 */
  esClient: Client;
  @Init()
  async initClient() {
    const esClient = new Client(this.esConfig);
    this.esClient = esClient;
  }
  /**信息检索-企业查询 */
  async company(keyword: string, skip: number = 0, limit: number = 10, others: object) {
    const index = 'company';
    const fields = ['industry', 'tags', 'name', 'pattern', 'scale', 'type', 'area', 'email', 'address', 'brief', 'products'];
    return this.oneIndexSearch(keyword, index, fields, skip, limit, others);
  }

  // /**信息检索-专家查询 */
  // async expert(keyword: string, skip: number = 0, limit: number = 10) {
  //   const index = 'expert';
  //   const fields = ['tags', 'name', 'area', 'industry', 'field', 'direction', 'work', 'education', 'title', 'brief', 'industry_type', 'work_type'];
  //   return this.oneIndexSearch(keyword, index, fields, skip, limit);
  // }
  /**信息检索-项目查询 */
  async project(keyword: string, skip: number = 0, limit: number = 10, others: object) {
    const index = 'project';
    const fields = ['tags', 'name', 'maturity', 'skill', 'type', 'area', 'field', 'cooperate', 'brief', 'main', 'progress', 'track_unit', 'source', 'industry'];
    return this.oneIndexSearch(keyword, index, fields, skip, limit, others);
  }
  /**信息检索-需求查询 */
  async demand(keyword: string, skip: number = 0, limit: number = 10, others: object) {
    const index = 'demand';
    const fields = ['industry', 'tags', 'name', 'field', 'urgent', 'method', 'area', 'brief', 'demand_status', 'company', 'company_brief', 'contacts', 'tec_name', 'question'];
    return this.oneIndexSearch(keyword, index, fields, skip, limit, others);
  }
  /**信息检索-供给查询 */
  async supply(keyword: string, skip: number = 0, limit: number = 10, others: object) {
    const index = 'supply';
    const fields = ['industry', 'tags', 'name', 'field', 'urgent', 'method', 'area', 'brief', 'demand_status'];
    return this.oneIndexSearch(keyword, index, fields, skip, limit, others);
  }
  /**信息检索-成果查询 */
  async achievement(keyword: string, skip: number = 0, limit: number = 10, others: object) {
    const index = 'achievement';
    const fields = ['industry', 'tags', 'name', 'patent', 'attribute', 'sell', 'mature', 'field', 'technology', 'area', 'brief', 'achievement_status', 'source'];
    return this.oneIndexSearch(keyword, index, fields, skip, limit, others);
  }

  /**
   * 信息检索查询
   * @param keyword 前端输入的关键词
   * @param index 针对哪个表查询
   * @param fields 哪些字段在查询范围内
   * @param skip 数据起始位置
   * @param limit 查询数量
   * @param params 模糊查询外的其他条件约束,都是用精确查询term
   * @returns
   */
  async oneIndexSearch(keyword: string, index: string, fields: Array<string>, skip: number, limit: number, params: object = {}) {
    const filter = [{ term: { is_use: '0' } }];
    const must: any = [];
    if (keyword) {
      const obj = {
        multi_match: {
          query: keyword,
          fields,
        },
      };
      must.push(obj)
    }

    if (Object.keys(params).length > 0) {
      for (const key in params) {
        const val = params[key];
        const obj = { match_phrase: { [key]: val } };
        must.push(obj);
      }
    }

    const result = await this.esClient.search({
      index,
      query: {
        bool: {
          must,
          filter,
        },
      },
      size: limit,
      from: skip,
      // sort: [{ _score: { order: 'desc' } }],
      // track_total_hits: true,
      // search_after: [skip],
    });
    const retrunData = this.dealResponses(result);
    return retrunData;
  }

  dealResponses(result) {
    const total = get(result, 'hits.total.value', 0);
    const hits = get(result, 'hits.hits', []);
    const list = [];
    for (const ol of hits) {
      const _score = get(ol, '_score', 0);
      const data = get(ol, '_source', {});
      let recommend = 0;
      /**推荐星数计算:超过5,就都是5颗星.未超过5的都向下取整 */
      if (_score >= 5) recommend = 5;
      else recommend = floor(_score);
      data._recommend = recommend;
      list.push(data);
    }
    return { total, data: list };
  }
}
