import { Client } from '@elastic/elasticsearch';
import { Provide, Config, Init } from '@midwayjs/core';
import * as dayjs from 'dayjs';
import { get } from 'lodash';
import { indices } from './elasticsearch/indices';
@Provide()
export class SyncService {
  @Config('elasticsearch')
  esConfig: object;
  /**es连接实例 */
  esClient: Client;
  @Init()
  async initClient() {
    const esClient = new Client(this.esConfig);
    this.esClient = esClient;
  }
  /**
   * 数据库创建&修改操作的同步
   * 1.根据mapping映射,处理下数据(e.g.:时间需要查看下是不是正常的时间格式)
   * @param index es索引(表名全小写)
   * @param data 数据
   * @param mapping 数据映射到es的设置
   */
  async upsert(index: string, data: object) {
    const hasIndex = await this.esClient.indices.exists({ index });
    if (!hasIndex) {
      // TODO: es没有找到索引,同步失败,需要写到日志库中
      console.error('upser no index');
      return;
    }
    const mapping = indices[index];
    if (!mapping) {
      // TODO: 没有找到mapping,同步失败,需要写到日志库中
      console.error('upser no mapping');
      return;
    }
    const afterDealData = this.dealData(data, mapping);
    const datasource = [afterDealData];
    const result = await this.esClient.helpers.bulk({
      datasource: datasource,
      onDocument(doc) {
        return [{ update: { _index: index, _id: get(doc, 'id') } }, { doc_as_upsert: true }];
      },
    });
    console.log('upsert finished');
    console.log(result);
    // TODO: es同步结果,记录到日志库中

    return 'ok';
  }

  /**
   * 数据库删除操作同步
   * @param index es索引(表名全小写)
   * @param data 数据
   */
  async delete(index: string, data: object) {
    const hasIndex = await this.esClient.indices.exists({ index });
    if (hasIndex) {
      // TODO: es没有找到索引,同步失败,需要写到日志库中
      console.error('delete no index');
      return;
    }
    const result = await this.esClient.helpers.bulk({
      datasource: [data],
      onDocument(doc) {
        return { delete: { _index: index, _id: get(doc, 'id') } };
      },
    });
    console.log('delete finished');
    console.log(result);
    // TODO: es同步结果,记录到日志库中
  }

  dealData(data: object, mapping: object) {
    const obj = {};
    for (const key in mapping) {
      const val = get(data, key);
      if (val) {
        // 检查是否是date类型,如果是date类型需要检查数据是否正确
        const dataType = mapping[key];
        if (dataType === 'date') {
          const isDate = dayjs(val).isValid();
          if (!isDate) continue;
        }
        let pk = key;
        obj[pk] = val;
      }
    }
    return obj;
  }
}
